﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  [Item("ComplexLeaf", "A leaf type that uses an arbitriary RegressionAlgorithm to create leaf models")]
  public class ComplexLeaf : ParameterizedNamedItem, ILeafType<IRegressionModel> {
    public const string RegressionParameterName = "Regression";
    public IValueParameter<IDataAnalysisAlgorithm<IRegressionProblem>> RegressionParameter {
      get { return Parameters[RegressionParameterName] as IValueParameter<IDataAnalysisAlgorithm<IRegressionProblem>>; }
    }
    public IDataAnalysisAlgorithm<IRegressionProblem> Regression {
      get { return RegressionParameter.Value; }
    }

    #region Constructors & Cloning
    [StorableConstructor]
    private ComplexLeaf(bool deserializing) : base(deserializing) { }
    private ComplexLeaf(ComplexLeaf original, Cloner cloner) : base(original, cloner) { }
    public ComplexLeaf() {
      var regression = new KernelRidgeRegression();
      Parameters.Add(new ValueParameter<IDataAnalysisAlgorithm<IRegressionProblem>>(RegressionParameterName, "The algorithm creating RegressionModels", regression));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ComplexLeaf(this, cloner);
    }
    #endregion

    #region IModelType
    public IRegressionModel BuildModel(IRegressionProblemData pd, IRandom random, CancellationToken cancellation, out int noParameters) {
      if (pd.Dataset.Rows < MinLeafSize(pd)) throw new ArgumentException("The number of training instances is too small to create a linear model");
      noParameters = pd.Dataset.Rows + 1;
      Regression.Problem = new RegressionProblem {ProblemData = pd};
      var res = M5StaticUtilities.RunSubAlgorithm(Regression, random.Next(), cancellation);
      var t = res.Select(x => x.Value).OfType<IRegressionSolution>().FirstOrDefault();
      if (t == null) throw new ArgumentException("No RegressionSolution was provided by the algorithm");
      return t.Model;
    }

    public int MinLeafSize(IRegressionProblemData pd) {
      return 3;
    }
    #endregion
  }
}